--[[
Copyright (c) 2012 John Glover

Permission is hereby granted, free of charge, to any person obtaining a copy
of newinst software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and newinst permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
]]

require 'slam'
require 'middleclass'
require 'Vector2d'
require 'Input'
local HC = require 'HardonCollider'
require 'tlfres'

game = { }

game.collider = { }
game.colliderLookup = { }
game.entities = { }
game.score = 0
game.scorecard = { }

game.delta = 0

game.width = 1920
game.height = 1080

game.paused = false
game.timescale = 1
game.lastTargetSpawn = 1

game.backArmAngle = 0
game.frontArmAngle = 0

game.xOffset = 20
game.yOffset = 20

game.started = false

game.menuInput = Input()
game.input = Input()
game.currentInput = game.menuInput

game.selectedResolution = 1
game.resolutions = 
{ 
	{ width = 1920, height = 1080, scale = 1 },
	{ width = 1600, height = 900, scale = 1600/1920 },
	{ width = 1366, height = 768, scale = 1366/1920 }, 
	{ width = 1280, height = 720, scale = 1280/1920 }, 
	{ width = 1024, height = 576, scale = 1024/1920 }, 
	{ width = 960, height = 540, scale = 960/1920 }, 
	{ width = 854, height = 480, scale = 854/1920 }, 
}

function handleCollision(dt, shape_a, shape_b, mtv_x, mtv_y)
	entity1, entity2 = game.colliderLookup[shape_a], game.colliderLookup[shape_b]
	local bulletEntity
	local targetEntity
	
	if entity1.name == 'bullet' then
		bulletEntity = entity1
	elseif entity1.name == 'can' or entity1.name == 'bottle' then
		targetEntity = entity1
	end
		
	if entity2.name == 'bullet' then
		bulletEntity = entity2
	elseif entity2.name == 'can' or entity2.name == 'bottle' then
	  targetEntity = entity2
	end
	
	if bulletEntity and targetEntity then
		bulletEntity.dead = true
		if targetEntity.name == 'can' then
			targetEntity.velocity = targetEntity.velocity:add(bulletEntity.velocity:scale(0.5))
			targetEntity.image = game.brokenCan
			game.score = game.score + 1
			table.insert(game.scorecard, { position = { }, image = game.deadCan })
			game.hitSound:play()
		elseif targetEntity.name == 'bottle' then
			targetEntity.dead = true
			game.spawnBrokenBottle(targetEntity)
			game.score = game.score + 2
			table.insert(game.scorecard, { position = { }, image = game.deadBottle })
			game.glassSound:play()
		end
	end
end

function game.spawnBrokenBottle(bottleEntity)
	local top = { 
		name = 'top',
		position = bottleEntity.position,
		velocity = bottleEntity.velocity:scale(0.5):add(Vector2d(0, math.random(200, 300))),
		angle = (math.random() - 0.5) * math.pi,
		scale = 1,
		image = game.brokenTop,
		xOffset = 19,
		yOffset = 64,
	}
		
	local bottom = { 
		name = 'bottom',
		position = bottleEntity.position,
		velocity = bottleEntity.velocity:add(Vector2d(0, math.random(200, 300))),
		angle = (math.random() - 0.5) * math.pi,
		scale = 1,
		image = game.brokenBottom,
		xOffset = 19,
		yOffset = 64,
	}
	
	table.insert(game.entities, top)
	table.insert(game.entities, bottom)
end

function endCollision(dt, shape_a, shape_b) end

function love.load()
	love.graphics.setFont(love.graphics.newFont(24))

	love.keypressed = game.menukeypressed
	love.draw = game.drawMenu
	love.update = game.updateMenu

	game.menuInput:bindKeystroke('up', function() 
			game.selectedResolution = game.clamp(game.selectedResolution - 1, 1, table.getn(game.resolutions))
		end, nil)
	game.menuInput:bindKeystroke('down', function()
			game.selectedResolution = game.clamp(game.selectedResolution + 1, 1, table.getn(game.resolutions))
		end, nil)	
	game.menuInput:bindKeystroke('return', function()
			local resolution = game.resolutions[game.selectedResolution]
      TLfres.setScreen({w = resolution.width, h = resolution.height, vsync = false, aa = 0}, 1920, false, false)
			game.load()
			game.currentInput = game.input
			love.update = game.update
			love.draw = game.draw
			love.keypressed = game.keypressed
		end, nil)
	game.menuInput:bindKeystroke('escape', function() love.event.push('quit') end, nil)
	game.menuInput:bindKeystroke('q', function() love.event.push('quit') end, nil)
end

function game.updateMenu() end

function game.load()
	love.graphics.setColorMode('replace')
	game.body = love.graphics.newImage('body.png')
	game.backArm = love.graphics.newImage('back-arm.png')
	game.backArmShooting = love.graphics.newImage('back-arm-flash.png')
	game.frontArm = love.graphics.newImage('front-arm.png')
	game.frontArmShooting = love.graphics.newImage('front-arm-flash.png')
	game.background = love.graphics.newImage('background.png')
	game.bullet = love.graphics.newImage('bullet.png')
	game.can = love.graphics.newImage('can2.png')
	game.brokenCan = love.graphics.newImage('broken-can.png')
	game.bottle = love.graphics.newImage('bottle.png')
	game.brokenTop = love.graphics.newImage('bottle-top.png')
	game.brokenBottom = love.graphics.newImage('bottle-bottom.png')
	game.crosshair = love.graphics.newImage('crosshair.png')
	game.startText = love.graphics.newImage('start.png')
	game.deadCan = love.graphics.newImage('deadCan.png')
	game.deadBottle = love.graphics.newImage('deadBottle.png')
	
	game.backArmCurrent = game.backArm
	game.frontArmCurrent = game.frontArm
	
	local width, height = game.width, game.height
	
	game.backShoulderVector = Vector2d(86 + game.xOffset, height - (512 + game.yOffset) + 168)
	game.frontShoulderVector = Vector2d(52 + game.xOffset, height - (512 + game.yOffset) + 151)
	
	game.insideArm = 2
	game.outsideArm = 1
	game.shotsFired = 0
	game.guns = { 
		{ name = 'outside', fired = 0, origin = Vector2d(317, -107) },
		{ name = 'inside', fired = 0, origin = Vector2d(302, -103) },
	}
	game.lastTargetSpawn = love.timer.getTime()
	
	game.verticalStart = height - (512 + game.yOffset)
	game.horizontalStart = game.xOffset

	game.collider = HC(100, handleCollision, endCollision)
	
	game.gunshot = love.audio.newSource('gunshot.ogg', 'static')
	game.hitSound = love.audio.newSource('hit.ogg', 'static')
	game.glassSound = love.audio.newSource('glass.ogg', 'static')
	game.music = love.audio.newSource('ugly.mp3', 'static') 
	
	love.mouse.setVisible(false)
	love.mouse.setGrab(true)

	game.input:bindKeystroke('pause', function( ) game.paused = not game.paused end, nil)
	game.input:bindKeystroke(' ', function( ) game.paused = not game.paused end, nil)
	game.input:bindMouse('l', game.shoot, nil, nil)
	game.input:bindKeystroke('escape', function() love.event.push('quit') end, nil)
	game.input:bindKeystroke('q', function() love.event.push('quit') end, nil)
end
		
function game.getMouseX()
  return love.mouse.getX() / love.graphics.getWidth() * game.width
end

function game.getMouseY()
  return love.mouse.getY() / love.graphics.getHeight() * game.height
end

function game.update(dt)
	local currentTime = love.timer.getTime()

	-- 160 is the length of the music in seconds.
	if game.started and currentTime - game.started > 160 then 
		game.started = nil 
		for _, v in pairs(game.entities) do
			if v.bounds then
				game.removeEntity(v)
			end
		end
		game.entities = { }
	end

	if not game.started or game.paused then return end
	dt = dt * game.timescale
	
  game.delta = game.delta + dt
  if game.delta > 1 / 60 then
    if math.random() < 0.002 * (love.timer.getTime() - game.lastTargetSpawn) then
      game.lastTargetSpawn = love.timer.getTime()
      game.newtarget()
    end
    delta = game.delta - 1 / 60
  end

  local mousePosition = Vector2d(game.getMouseX(), game.getMouseY())
	game.backArmAngle = mousePosition:subtract(game.backShoulderVector):angle()
	game.frontArmAngle = mousePosition:subtract(game.frontShoulderVector):angle()

	if game.guns[2].fired + 0.5 < currentTime then
		game.backArmCurrent = game.backArm
	else
		game.backArmCurrent = game.backArmShooting
	end
	
	if game.guns[1].fired + 0.5 < currentTime then
		game.frontArmCurrent = game.frontArm
	else
		game.frontArmCurrent = game.frontArmShooting
	end
	
	local livingEntities = { }
	for _, v in pairs(game.entities) do
		v.position = v.velocity:scale(dt):add(v.position)
		if game.outOfBounds(v.position) then
			v.dead = true
		else
			if v.rotate then
				v.angle = v.angle + dt
				if v.bounds then
					v.bounds:setRotation(v.angle)
				end
			end
			if v.bounds then
				local temp = v.position
				v.bounds:moveTo(temp.x, temp.y)
				v.bounds:setRotation(v.angle)
			end
		end
		
		if v.bounds and v.dead then
			game.removeEntity(v)
		else
			table.insert(livingEntities, v)
		end
	end
	game.entities = livingEntities
	
	game.collider:update(dt)
end

function game.outOfBounds(vector)
	if vector.y > game.height + 200 
		or vector.x > game.width + 200 
		or vector.y < -200 or vector.x < -200 then
			return true
	end
	return false
end

function game.removeEntity(entity)
	game.colliderLookup[entity.bounds] = nil
	game.collider:remove(entity.bounds)
end
 
function game.draw()
  TLfres.transform()

	love.graphics.draw(game.background, 0, 0, 0) --, scale, scale)
	love.graphics.draw(game.backArmCurrent, game.horizontalStart + 86, game.verticalStart + 168, game.backArmAngle, 1, 1, 86, 168)
	love.graphics.draw(game.body, game.horizontalStart, game.verticalStart, 0)
	love.graphics.draw(game.frontArmCurrent, game.horizontalStart + 52, game.verticalStart + 151, game.frontArmAngle, 1, 1, 52, 151)
	
	for _, v in pairs(game.entities) do
		love.graphics.draw(v.image, v.position.x, v.position.y, v.angle, v.scale, v.scale, v.xOffset, v.yOffset)
		--if v.bounds then
		--	v.bounds:draw()
		--end
	end
	
	love.graphics.draw(game.crosshair, game.getMouseX() - game.crosshair:getWidth() / 2, game.getMouseY() - game.crosshair:getHeight() / 2)
	
	game.drawScore()
	
	if not game.started then
		local width, height = game.width, game.height
		
		love.graphics.draw(game.startText, width / 2, height / 2, 0, 1, 1, 384, 48)
	end

  TLfres.letterbox(16, 9)
end

function game.drawScore( )
	local width, height = game.width, game.height
	
	local placement = Vector2d(0, height - 32)

	for _, v in pairs(game.scorecard) do
		love.graphics.draw(v.image, placement.x, placement.y, 0, 0.5, 0.5, 16, 32)
		placement.x = placement.x + 32
		if placement.x > width - 32 then
		  placement.x = 0
			placement.y = placement.y - 64
		end
	end
end

function game.drawMenu( )
	local startValue = 10
	local stepValue = 28
	local currentHeight = 10

	br, bg, bb, ba = love.graphics.getBackgroundColor()
	love.graphics.setBackgroundColor(0, 0, 0, 255)
	for i = 1, table.getn(game.resolutions) do
		local resolution = game.resolutions[i].width .. 'x' .. game.resolutions[i].height
		r, g, b, a = love.graphics.getColor()
		if i == game.selectedResolution then
			love.graphics.setColor(0, 255, 0, 255)
		else
			love.graphics.setColor(255, 255, 255, 255)
		end
		love.graphics.print(resolution, startValue, currentHeight)
		love.graphics.setColor(r, g, b, a)
		currentHeight = currentHeight + stepValue
	end
	love.graphics.setBackgroundColor(br, bg, bb)
end

function game.clamp(number, min, max)
	return math.min(math.max(number, min), max)
end

function game.menukeypressed(key, unicode)
	game.currentInput:keypressed(key, unicode)
end

function game.keypressed(key, unicode)
	if not game.started then
		game.started = love.timer.getTime()
		love.audio.play(game.music)
		game.scorecard = { }
		game.shotsFired = 0
		return
	end
	game.currentInput:keypressed(key, unicode)
end

function love.keyreleased(key, unicode)
	game.currentInput:keyreleased(key, unicode)
end

function love.mousepressed(x, y, button)
	if not game.started then
		return
	end
	game.currentInput:mousepressed(x, y, button)
end

function love.mousereleased(x, y, button)
	game.currentInput:mousereleased(x, y, button)
end

function game.shoot()
	print('shooting')
	local currentGun = game.guns[(game.shotsFired % 2) + 1]
	game.shotsFired = game.shotsFired + 1
	currentGun.fired = love.timer.getTime()
	game.gunshot:play()
	
	local angle = game.frontArmAngle
	local vector = game.frontShoulderVector
	if currentGun.name == 'inside' then 
		angle = game.backArmAngle 
		vector = game.backShoulderVector
	end
	
	local position = currentGun.origin:rotate(angle):add(vector)
	local velocity = Vector2d(1000, 0):rotate(angle + math.pi / 24.5)
	local bounds = game.collider:addRectangle(position.x, position.y, 32, 16)
	bounds:rotate(angle)
	
	local bullet = { 
	  name = 'bullet',
		position = position,
		velocity = velocity,
		angle = angle + math.pi / 24.5,
		scale = 0.25,
		image = game.bullet,
		xOffset = 16,
		yOffset = 8,
		bounds = bounds,
	}
	
	table.insert(game.entities, bullet)
	game.colliderLookup[bounds] = bullet
end

function game.newtarget()
	if math.random() < 0.7 then
		game.spawnCan()
	else
		game.spawnBottle()
	end
end

function game.spawnBottle()
	local width, height = game.width, game.height
	print('spawning bottle')
	local position = Vector2d(width, math.random(60, height - 60))
	local velocity = Vector2d(math.random(0, 100) - 300, math.random(0, 50) -25)
	local bounds = game.collider:addRectangle(position.x, position.y, 38, 128)
	
	target = {
		name = 'bottle',
		position = position,
		velocity = velocity,
		angle = 0,
		scale = 1,
		image = game.bottle,
		xOffset = 19,
		yOffset = 64,
		bounds = bounds,
	}
	
	table.insert(game.entities, target)
	game.colliderLookup[bounds] = target
end

function game.spawnCan()
	local width, height = game.width, game.height
	print('spawning can')
	local position = Vector2d(width, math.random(60, height - 60))
	local velocity = Vector2d(math.random(0, 100) - 200, math.random(0, 50) -25)
	local bounds = game.collider:addRectangle(position.x, position.y, 64, 128)
	
	target = {
		name = 'can',
		position = position,
		velocity = velocity,
		angle = 0,
		scale = 1,
		image = game.can,
		xOffset = 32,
		yOffset = 64,
		rotate = true,
		bounds = bounds,
	}
	
	table.insert(game.entities, target)
	game.colliderLookup[bounds] = target
end